@@ -1,6 +1,6 @@
 /*----------------------------------------------------------------------------
  * avs2bdnxml - Generates BluRay subtitle stuff from RGBA AviSynth scripts
- * Copyright (C) 2008-2009 Arne Bochem <avs2bdnxml at ps-auxw de>
+ * Copyright (C) 2008-2010 Arne Bochem <avs2bdnxml at ps-auxw de>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,6 +15,27 @@
  * You should have received a copy of the GNU General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *----------------------------------------------------------------------------
+ * Version 2.05
+ *   - Fix crash when using XML and SUP output
+ *
+ * Version 2.04
+ *   - Minor fixes
+ *   - Code cleanup regarding linked lists
+ *   - Add new minimal window finding code
+ *   - Fix PGS output more, especially WDS handling
+ *   - Fix old bug in RLE encoder
+ *
+ * Version 2.03
+ *   - Output some statistics about previous epoch in pgsparse in end packet
+ *   - Fix unnecessary epoch splits due to bad buffer calculation and more.
+ *   - Some small fixes
+ *
+ * Version 2.02
+ *   - It is now possible to use -o twice, with different output formats
+ *   - Corrections to SUP timestamp calculation
+ *   - PGS decode buffer should contain palettized image data, not RGBA.
+ *     Buffer usage calculation fixed.
+ *
  * Version 2.01
  *   - Restrict number of composition objects within an epoch to 64
  *
@@ -99,6 +120,7 @@
  *   - Initial release. Seems to work.
  *----------------------------------------------------------------------------
  * Thanks to:
+ *   - hamletiii on Doom9 for a steady flow of feedback.
  *   - Daiz in #x264 for inspiration, testing, and bug reports.
  *   - Loren Merritt and Laurent Aimar for writing x264's AVS reader code.
  *----------------------------------------------------------------------------*/
@@ -114,6 +136,7 @@
 #include "auto_split.h"
 #include "palletize.h"
 #include "sup.h"
+#include "abstract_lists.h"
 
 /* AVIS input code taken from muxers.c from the x264 project (GPLv2 or later).
  * Authors: Laurent Aimar <fenrir@via.ecp.fr>
@@ -125,6 +148,7 @@
 #include <vfw.h>
 #endif
 
+#ifndef LINUX
 static int64_t gcd( int64_t a, int64_t b )
 {
     while (1)
@@ -136,6 +160,7 @@ static int64_t gcd( int64_t a, int64_t b )
         b = c;
     }
 }
+#endif
 
 typedef struct {
 #ifndef LINUX
@@ -206,7 +231,7 @@ int open_file_avis( char *psz_filename, avis_input_t **p_handle, stream_info_t *
     p_param->i_height = 1080;
     p_param->i_fps_den = 30000;
     p_param->i_fps_num = 1001;
-    *p_handle = fopen(psz_filename, "r");
+    *p_handle = (avis_input_t *)fopen(psz_filename, "r");
 		return 0;
 #endif
 }
@@ -236,7 +261,7 @@ int read_frame_avis( char *p_pic, avis_input_t *handle, int i_frame )
 
     return 0;
 #else
-    fread(p_pic, 4, 1920*1080, handle);
+    fread(p_pic, 4, 1920*1080, (FILE *)handle);
     return 0;
 #endif
 }
@@ -250,7 +275,7 @@ int close_file_avis( avis_input_t *handle )
     free(h);
     return 0;
 #else
-    fclose(handle);
+    fclose((FILE *)handle);
     return 0;
 #endif
 }
@@ -564,7 +589,10 @@ int detect_sse2 ()
 	unsigned int eax, ebx, ecx, edx;
 
 	/* FIXME: SSE2 broken with new GCC. */
-	/* return 0; */
+#if __GNUC__ != 3
+#warning "Disabling SSE2 assembly."
+	return 0;
+#endif
 
 	if (detection != -1)
 		return detection;
@@ -697,44 +725,30 @@ int is_extension(char *filename, char *check_ext)
 	return 0;
 }
 
-struct event_list_node_s
+typedef struct event_s
 {
 	int image_number;
 	int start_frame;
 	int end_frame;
 	int graphics;
 	crop_t c[2];
-	struct event_list_node_s *next;
-};
+} event_t;
 
-typedef struct
-{
-	struct event_list_node_s *start;
-	struct event_list_node_s *end;
-} event_list_head_t;
+STATIC_LIST(event, event_t)
 
-void add_event_xml_real (event_list_head_t *events, int image, int start, int end, int graphics, crop_t *crops)
+void add_event_xml_real (event_list_t *events, int image, int start, int end, int graphics, crop_t *crops)
 {
-	struct event_list_node_s *new = calloc(1, sizeof(struct event_list_node_s));
+	event_t *new = calloc(1, sizeof(event_t));
 	new->image_number = image;
 	new->start_frame = start;
 	new->end_frame = end;
 	new->graphics = graphics;
 	new->c[0] = crops[0];
 	new->c[1] = crops[1];
-	if (events->start == NULL)
-	{
-		events->start = new;
-		events->end = new;
-	}
-	else
-	{
-		(events->end)->next = new;
-		events->end = new;
-	}
+	event_list_insert_after(events, new);
 }
 
-void add_event_xml (event_list_head_t *events, int split_at, int min_split, int start, int end, int graphics, crop_t *crops)
+void add_event_xml (event_list_t *events, int split_at, int min_split, int start, int end, int graphics, crop_t *crops)
 {
 	int image = start;
 	int d = end - start;
@@ -756,7 +770,6 @@ void add_event_xml (event_list_head_t *events, int split_at, int min_split, int
 
 void write_sup_wrapper (sup_writer_t *sw, uint8_t *im, int num_crop, crop_t *crops, uint32_t *pal, int start, int end, int split_at, int min_split)
 {
-	int image = start;
 	int d = end - start;
 
 	if (!split_at)
@@ -804,7 +817,9 @@ int main (int argc, char *argv[])
 	char *language = "und";
 	char *video_format = "1080p";
 	char *frame_rate = "23.976";
-	char *out_filename = NULL;
+	char *out_filename[2] = {NULL, NULL};
+	char *sup_output_fn = NULL;
+	char *xml_output_fn = NULL;
 	char *x_offset = "0";
 	char *y_offset = "0";
 	char *buffer_optimize = "0";
@@ -821,6 +836,7 @@ int main (int argc, char *argv[])
 	crop_t crops[2];
 	pic_t pic;
 	uint32_t *pal = NULL;
+	int out_filename_idx = 0;
 	int have_fps = 0;
 	int n_crop = 1;
 	int split_at = 0;
@@ -846,11 +862,11 @@ int main (int argc, char *argv[])
 	int fps_num = 25, fps_den = 1;
 	int sup_output = 0;
 	int xml_output = 0;
-	sup_writer_t *sw;
+	sup_writer_t *sw = NULL;
 	avis_input_t *avis_hnd;
 	stream_info_t *s_info = malloc(sizeof(stream_info_t));
-	event_list_head_t *events = malloc(sizeof(event_list_head_t));
-	struct event_list_node_s *node;
+	event_list_t *events = event_list_new();
+	event_t *event;
 	FILE *fh;
 
 	/* Get args */
@@ -887,7 +903,13 @@ int main (int argc, char *argv[])
 			switch (c)
 			{
 				case 'o':
-					out_filename = optarg;
+					if (out_filename_idx < 2)
+						out_filename[out_filename_idx++] = optarg;
+					else
+					{
+						fprintf(stderr, "No more than two output filenames allowed.\nIf more than one is used, the other must have a\ndifferent output format.\n");
+						exit(0);
+					}
 					break;
 				case 'j':
 					seek_string = optarg;
@@ -951,28 +973,36 @@ int main (int argc, char *argv[])
 		print_usage();
 		return 0;
 	}
-	if (out_filename == NULL)
+	if (out_filename[0] == NULL)
 	{
 		print_usage();
 		return 0;
 	}
 
 	/* Get target output format */
-	if (is_extension(out_filename, "xml"))
+	for (i = 0; i < out_filename_idx; i++)
 	{
-		xml_output = 1;
-		sup_output = 0;
+		if (is_extension(out_filename[i], "xml"))
+		{
+			xml_output_fn = out_filename[i];
+			xml_output++;
+		}
+		else if (is_extension(out_filename[i], "sup") || is_extension(out_filename[i], "pgs"))
+		{
+			sup_output_fn = out_filename[i];
+			sup_output++;
+			pal_png = 1;
+		}
+		else
+		{
+			fprintf(stderr, "Output file extension must be \".xml\", \".sup\" or \".pgs\".\n");
+			return 1;
+		}
 	}
-	else if (is_extension(out_filename, "sup") || is_extension(out_filename, "pgs"))
+	if (sup_output > 1 || xml_output > 1)
 	{
-		sup_output = 1;
-		xml_output = 0;
-		pal_png = 1;
-	}
-	else
-	{
-		fprintf(stderr, "Output file extension must be \".xml\", \".sup\" or \".pgs\".\n");
-		return 1;
+		fprintf(stderr, "If more than one output filename is used, the other must have a\ndifferent output format.\n");
+		exit(0);
 	}
 
 	/* Set X and Y offsets, and split value */
@@ -1032,9 +1062,9 @@ int main (int argc, char *argv[])
 	}
 
 	/* Align buffers */
-	in_img  = in_img + (16 - ((int)in_img % 16));
-	old_img = old_img + (16 - ((int)old_img % 16));
-	out_buf = out_buf + (16 - ((int)out_buf % 16));
+	in_img  = in_img + (short)(16 - ((long)in_img % 16));
+	old_img = old_img + (short)(16 - ((long)old_img % 16));
+	out_buf = out_buf + (short)(16 - ((long)out_buf % 16));
 
 	/* Set up buffer (non-)optimization */
 	buffer_opt = atoi(buffer_optimize);
@@ -1075,14 +1105,8 @@ int main (int argc, char *argv[])
 	}
 
 	/* Open SUP writer, if applicable */
-	sw = new_sup_writer(out_filename, pic.w, pic.h, fps_num, fps_den);
-
-	/* Initialize events */
-	events->start = NULL;
-	events->end = NULL;
-
-	if (detect_sse2())
-		asm ("emms\n");
+	if (sup_output)
+		sw = new_sup_writer(sup_output_fn, pic.w, pic.h, fps_num, fps_den);
 
 	/* Process frames */
 	for (i = init_frame; i < last_frame; i++)
@@ -1122,11 +1146,12 @@ int main (int argc, char *argv[])
 			if (sup_output)
 			{
 				assert(pal != NULL);
-				write_sup_wrapper(sw, out_buf, n_crop, crops, pal, start_frame, i, split_at, min_split);
-				free(pal);
+				write_sup_wrapper(sw, (uint8_t *)out_buf, n_crop, crops, pal, start_frame, i, split_at, min_split);
+				if (!xml_output)
+					free(pal);
 				pal = NULL;
 			}
-			else if (xml_output)
+			if (xml_output)
 				add_event_xml(events, split_at, min_split, start_frame, i, n_crop, crops);
 			end_frame = i;
 			have_line = 0;
@@ -1155,12 +1180,12 @@ int main (int argc, char *argv[])
 			crops[0].h = pic.h;
 			auto_crop(pic, &(crops[0]));
 		}
-		if (pal_png)
+		if ((pal_png || sup_output) && pal == NULL)
 			pal = palletize(out_buf, s_info->i_width, s_info->i_height);
 		if (xml_output)
 			for (j = 0; j < n_crop; j++)
-				write_png(start_frame, out_buf, s_info->i_width, s_info->i_height, j, pal, crops[j]);
-		if (pal_png && !sup_output)
+				write_png(start_frame, (uint8_t *)out_buf, s_info->i_width, s_info->i_height, j, pal, crops[j]);
+		if (pal_png && xml_output && !sup_output)
 		{
 			free(pal);
 			pal = NULL;
@@ -1183,12 +1208,17 @@ int main (int argc, char *argv[])
 		if (sup_output)
 		{
 			assert(pal != NULL);
-			write_sup_wrapper(sw, out_buf, n_crop, crops, pal, start_frame, i - 1, split_at, min_split);
+			write_sup_wrapper(sw, (uint8_t *)out_buf, n_crop, crops, pal, start_frame, i - 1, split_at, min_split);
+			if (!xml_output)
+				free(pal);
+			pal = NULL;
+		}
+		if (xml_output)
+		{
+			add_event_xml(events, split_at, min_split, start_frame, i - 1, n_crop, crops);
 			free(pal);
 			pal = NULL;
 		}
-		else if (xml_output)
-			add_event_xml(events, split_at, min_split, start_frame, i - 1, n_crop, crops);
 		auto_cut = 1;
 		end_frame = i - 1;
 	}
@@ -1207,15 +1237,12 @@ int main (int argc, char *argv[])
 			return 1;
 		}
 
-		if (detect_sse2())
-			asm ("emms\n");
-
 		/* Initialize timecode buffers */
 		intc_buf = calloc(12, 1);
 		outtc_buf = calloc(12, 1);
 
 		/* Creating output file */
-		if ((fh = fopen(out_filename, "w")) == 0)
+		if ((fh = fopen(xml_output_fn, "w")) == 0)
 		{
 			perror("Error opening output XML file");
 			return 1;
@@ -1240,28 +1267,28 @@ int main (int argc, char *argv[])
 			"<Events>\n", intc_buf, outtc_buf, num_of_events);
 
 		/* Write XML events */
-		if (events->start != NULL)
+		if (!event_list_empty(events))
 		{
-			node = events->start;
+			event = event_list_first(events);
 			do
 			{
-				mk_timecode(node->start_frame, fps, intc_buf);
-				mk_timecode(node->end_frame, fps, outtc_buf);
+				mk_timecode(event->start_frame, fps, intc_buf);
+				mk_timecode(event->end_frame, fps, outtc_buf);
 				
-				if (auto_cut && node->end_frame == frames - 1)
+				if (auto_cut && event->end_frame == frames - 1)
 				{
-					mk_timecode(node->end_frame + 1, fps, outtc_buf);
+					mk_timecode(event->end_frame + 1, fps, outtc_buf);
 				}
 				
 				fprintf(fh, "<Event Forced=\"False\" InTC=\"%s\" OutTC=\"%s\">\n", intc_buf, outtc_buf);
-				for (i = 0; i < node->graphics; i++)
+				for (i = 0; i < event->graphics; i++)
 				{
-					fprintf(fh, "<Graphic Width=\"%d\" Height=\"%d\" X=\"%d\" Y=\"%d\">%08d_%d.png</Graphic>\n", node->c[i].w, node->c[i].h, xo + node->c[i].x, yo + node->c[i].y, node->image_number, i);
+					fprintf(fh, "<Graphic Width=\"%d\" Height=\"%d\" X=\"%d\" Y=\"%d\">%08d_%d.png</Graphic>\n", event->c[i].w, event->c[i].h, xo + event->c[i].x, yo + event->c[i].y, event->image_number, i);
 				}
 				fprintf(fh, "</Event>\n");
-				node = node->next;
+				event = event_list_next(events);
 			}
-			while (node != NULL);
+			while (event != NULL);
 		}
 
 		/* Write XML footer */
