@@ -19,9 +19,11 @@
 #include <stdint.h>
 #include <stdlib.h>
 #include <stdio.h>
+#include <string.h>
 #include <math.h>
 #include "auto_split.h"
 #include "sup.h"
+#include "abstract_lists.h"
 
 #ifdef BE_ARCH
 #define SWAP32(x) (x)
@@ -37,6 +39,10 @@
 #endif
 #endif
 
+#ifndef DEBUG
+#define DEBUG 0
+#endif
+
 static int count (uint8_t *im, int x, int w, int col)
 {
 	int c = 0;
@@ -49,13 +55,14 @@ static int count (uint8_t *im, int x, int w, int col)
 #define PUSH(x) {*(b++)=(x);(*len)++;}
 #define FLAG_COLOR 0x80
 #define FLAG_LONG 0x40
-static uint8_t *rl_encode (uint8_t *im, int w, int h, crop_t crop, int *len)
+static uint8_t *rl_encode (uint8_t *im, int w, int h, rect_t crop, int *len)
 {
 	uint8_t *b = calloc(crop.w * crop.h, 4); /* Over-allocation */
 	uint8_t *rle = b;
 	uint8_t col, o;
 	int x, y, c;
 
+	*len = 0;
 	for (y = crop.y; y < crop.y + crop.h && y < h; y++)
 	{
 		for (x = crop.x; x < crop.x + crop.w && x < w; x += c)
@@ -85,7 +92,7 @@ static uint8_t *rl_encode (uint8_t *im, int w, int h, crop_t crop, int *len)
 				o |= FLAG_LONG;
 				o |= c >> 8;
 				PUSH(o);
-				PUSH(c & 0xFF)
+				PUSH(c & 0xFF);
 			}
 			else
 				PUSH(o | c);
@@ -156,7 +163,7 @@ void conv_sup_pcs_start (sup_pcs_start_t *pcss)
 typedef struct sup_pcs_start_obj_s
 {
 	uint16_t picture;
-	uint8_t object; /* 0 */
+	uint8_t window;
 	uint8_t forced; /* forced ? 64 : 0 */
 	uint16_t x_off;
 	uint16_t y_off;
@@ -187,12 +194,12 @@ static void write_pcs_start (FILE *fh, int start_time, int dts, int follower, in
 	fwrite(&pcss, sizeof(pcss), 1, fh);
 }
 
-static void write_pcs_start_obj (FILE *fh, int picture, int object, int x_off, int y_off)
+static void write_pcs_start_obj (FILE *fh, int picture, int window, int x_off, int y_off)
 {
 	sup_pcs_start_obj_t pcsso;
 
 	pcsso.picture = picture;
-	pcsso.object = object;
+	pcsso.window = window;
 	pcsso.forced = 0;
 	pcsso.x_off = x_off;
 	pcsso.y_off = y_off;
@@ -203,7 +210,7 @@ static void write_pcs_start_obj (FILE *fh, int picture, int object, int x_off, i
 
 typedef struct sup_wds_s
 {
-	uint8_t objects; /* 1 or 2 */
+	uint8_t windows; /* 1 or 2 */
 } __attribute ((packed)) sup_wds_t;
 
 void conv_sup_wds (sup_wds_t *wds)
@@ -213,7 +220,7 @@ void conv_sup_wds (sup_wds_t *wds)
 
 typedef struct sup_wds_obj_s
 {
-	uint8_t object; /* 0 or 1 */
+	uint8_t window; /* 0 or 1 */
 	uint16_t x_off;
 	uint16_t y_off;
 	uint16_t width;
@@ -228,23 +235,23 @@ void conv_sup_wds_obj (sup_wds_obj_t *wdso)
 	wdso->height = SWAP16(wdso->height);
 }
 
-static void write_wds (FILE *fh, int timestamp, int dts, int objects)
+static void write_wds (FILE *fh, int timestamp, int dts, int windows)
 {
 	sup_wds_t wds;
 
-	write_header(fh, timestamp, dts, 23, sizeof(wds) + objects * sizeof(sup_wds_obj_t));
+	write_header(fh, timestamp, dts, 23, sizeof(wds) + windows * sizeof(sup_wds_obj_t));
 
-	wds.objects = objects;
+	wds.windows = windows;
 
 	conv_sup_wds(&wds);
 	fwrite(&wds, sizeof(wds), 1, fh);
 }
 
-static void write_wds_obj (FILE *fh, int object, int w, int h, int x_off, int y_off)
+static void write_wds_obj (FILE *fh, int window, int w, int h, int x_off, int y_off)
 {
 	sup_wds_obj_t wdso;
 
-	wdso.object = object;
+	wdso.window = window;
 	wdso.x_off = x_off;
 	wdso.y_off = y_off;
 	wdso.width = w;
@@ -285,17 +292,17 @@ static uint8_t get_v (uint32_t c, int s)
 
 typedef struct sup_palette_s
 {
-	uint16_t picture;
+	uint16_t palette;
 } __attribute ((packed)) sup_palette_t;
 
 void conv_sup_palette (sup_palette_t *p)
 {
-	p->picture = SWAP16(p->picture);
+	p->palette = SWAP16(p->palette);
 }
 
 /* Colorspace = 1 for 480p/576p, 0 otherwise */
 #define PUT(x) { t = (uint8_t)(x); fwrite(&t, 1, 1, fh); }
-static void write_palette (FILE *fh, int dts, int picture, uint32_t *pal, int colorspace)
+static void write_palette (FILE *fh, int dts, int palette, uint32_t *pal, int colorspace)
 {
 	sup_palette_t p;
 	int entries = 1, i;
@@ -305,7 +312,7 @@ static void write_palette (FILE *fh, int dts, int picture, uint32_t *pal, int co
 		entries++;
 	write_header(fh, dts, 0, 20, sizeof(p) + entries * 5);
 
-	p.picture = picture;
+	p.palette = palette;
 	conv_sup_palette(&p);
 	fwrite(&p, sizeof(p), 1, fh);
 
@@ -459,7 +466,6 @@ static int get_id (int fps_num, int fps_den)
 sup_writer_t *new_sup_writer (char *filename, int im_w, int im_h, int fps_num, int fps_den)
 {
 	sup_writer_t *sw = malloc(sizeof(sup_writer_t));
-	int i;
 
 	if ((sw->fh = fopen(filename, "wb")) == NULL)
 	{
@@ -481,156 +487,338 @@ sup_writer_t *new_sup_writer (char *filename, int im_w, int im_h, int fps_num, i
 	sw->fps_id = get_id(fps_num, fps_den);
 	sw->comp_num = 0;
 	sw->end = -2;
-	sw->picture = 0;
+	sw->follower_end = -2;
+	sw->objects = 0;
+	sw->palettes = 0;
 	sw->buffer = 0;
-	sw->last.end_ts = -2;
-	sw->last.window_ts = 0;
-	sw->last.num_crop = 0;
-	for (i = 0; i < 2; i++)
-	{
-		sw->last.crops[i].x = 0;
-		sw->last.crops[i].y = 0;
-		sw->last.crops[i].w = 0;
-		sw->last.crops[i].h = 0;
-	}
+	sw->palette_offset = 0;
+	sw->picture_offset = 0;
+	sw->last_end_ts = 0;
+	sw->last_window_ts = 0;
+	sw->window_num = 0;
+	sw->sil = si_list_new();
+
+	memset(sw->windows, 0, 2 * sizeof(rect_t));
 
 	return sw;
 }
 
-void end_composition (sup_writer_t *sw)
+void destroy_si (subtitle_info_t *si)
 {
-	int ts, dts;
 	int i;
 
-	/* Only write end marker if anything was written */
-	if (!sw->non_new)
-		return;
+	for (i = 0; i < si->num_crop; i++)
+		free(si->rle[i]);
 
-	/* New composition */
-	(sw->comp_num)++;
-	sw->buffer = 0;
-
-	/* Reset picture count, new buffer. */
-	sw->picture = 0;
-
-	/* Write PCSE */
-	dts = sw->last.end_ts - sw->last.window_ts - 1;
-	write_pcs_end(sw->fh, sw->last.end_ts, dts, sw->im_w, sw->im_h, sw->fps_id, sw->comp_num); /* Ok! */
-
-	/* Write WDS */
-	ts = sw->last.end_ts - sw->last.window_ts;
-	write_wds(sw->fh, ts, dts, sw->last.num_crop);
-	for (i = 0; i < sw->last.num_crop; i++)
-		write_wds_obj(sw->fh, i, sw->last.crops[i].w, sw->last.crops[i].h, sw->last.crops[i].x, sw->last.crops[i].y);
-
-	/* Write marker */
-	write_marker(sw->fh, dts);
-
-	/* Adjust composition number */
-	(sw->comp_num)++;
+	free(si);
 }
 
-void close_sup_writer (sup_writer_t *sw)
+void write_subtitle (sup_writer_t *sw, uint8_t **rle, int *rle_len, int num_crop, rect_t *crops, uint32_t *pal, int start, int end, int new_composition)
 {
-	end_composition(sw);
-	fclose(sw->fh);
-	free(sw);
-}
-
-void write_sup (sup_writer_t *sw, uint8_t *im, int num_crop, crop_t *crops, uint32_t *pal, int start, int end)
-{
-	int rle_len = 0;
-	uint8_t *rle;
-	int frame_ts, window_ts;
-	int dts;
-	int start_ts, end_ts, ts;
+	uint32_t frame_ts, window_ts, decode_ts;
+	uint32_t window_ts_list[2], decode_ts_list[2];
+	uint32_t later_window;
+	int in_window[2];
+	uint32_t dts;
+	uint32_t start_ts, end_ts, ts;
 	int follower = 0;
-	int im_ts, im_dts;
-	int buffer_increase;
-	int i;
+	uint32_t im_ts = 0;
+	int i, j;
 	double tick_fac = 90000;
+
+	/* For stuff following frame by frame (endts = startts):
+	 *   - Write: PCSS, WDS, PAL, IMG, MARK, [PCSS, WDS, PAL, IMG, MARK...], PCSE, WDS, MARK
+	 *
+	 * Behaviour in reference file:
+	 *   During an epoch of followers, if number of objects/position/size changes, increase
+	 *   composition by one, increase all further picture values by previous number of
+	 *   composition objects, increase palette by one. Object id in PCSS is 1 for all.
+	 *   Do some stuff to WDS too. Keep it to first value, if stuff still fits in the window?
+	 *   It seems WDS has to stay constant withing a single composition. What a pain.
+	 */
 	tick_fac *= ((double)sw->fps_den) / ((double)sw->fps_num);
 
 	start_ts = (int)floor((double)start * tick_fac + 0.5);
 	end_ts = (int)floor((double)end * tick_fac + 0.5);
 
-	buffer_increase = 0;
-	for (i = 0; i < num_crop; i++)
-		buffer_increase += crops[i].w * crops[i].h * 4 + 16;
-	if (sw->non_new && ((start_ts > sw->end) || (sw->buffer + buffer_increase >= 4 * 1024 * 1024) || (sw->picture + num_crop > 64)))
-		end_composition(sw);
-	if (sw->non_new && ((start_ts == sw->end) || (start_ts == sw->end + 1)))
-		follower = 1;
-	sw->non_new = 1;
-	sw->end = end_ts;
-
-	/* For stuff following frame by frame (endts = startts):
-	 *   - Write: PCSS, WDS, PAL, IMG, MARK, [PCSS, WDS, PAL, IMG, MARK...], PCSE, WDS, MARK
-	 * For stuff following closely (endts = startts or endts+1 = startts):
-	 *   - Set: pcss.follower = 0x40
-	 */
-
 	/* Calculate some timestamps/modifiers */
 	frame_ts = (sw->im_w * sw->im_h * 9 + 3199) / 3200;
 	window_ts = 0;
+
+	if (sw->non_new && ((start == sw->follower_end) || (start == sw->follower_end + 1)))
+		follower = 1;
+	sw->follower_end = end;
+
+	for (i = 0; i < sw->window_num; i++)
+	{
+		window_ts_list[i] = (sw->windows[i].w * sw->windows[i].h * 9 + 3199) / 3200;
+		window_ts += window_ts_list[i];
+	}
+	decode_ts = 0;
 	for (i = 0; i < num_crop; i++)
-		window_ts += (crops[i].w * crops[i].h * 9 + 3199) / 3200;
-	dts = start_ts - (frame_ts + window_ts);
+	{
+		decode_ts_list[i] = (crops[i].w * crops[i].h * 9 + 1599) / 1600;
+		decode_ts += decode_ts_list[i];
+	}
+
+	if (sw->window_num > 1)
+		later_window = window_ts_list[1];
+	else
+		later_window = window_ts_list[0];
+
+	/* Calculate dts */
+	if (num_crop == 1)
+	{
+		if (new_composition)
+			dts = start_ts - frame_ts - window_ts;
+		else
+			dts = start_ts - window_ts - decode_ts;
+	}
+	else
+	{
+		if (new_composition)
+			dts = start_ts - frame_ts - window_ts;
+		else
+			dts = start_ts - decode_ts - later_window;
+	}
+
+	/* Determine windows */
+	for (i = 0; i < num_crop; i++)
+		for (j = 0; j < sw->window_num; j++)
+			if (crops[i].x >= sw->windows[j].x && crops[i].x + crops[i].w <= sw->windows[j].x + sw->windows[j].w && crops[i].y >= sw->windows[j].y && crops[i].y + crops[i].h <= sw->windows[j].y + sw->windows[j].h)
+			{
+				in_window[i] = j;
+				break;
+			}
 
 	/* Write PCSS */
 	write_pcs_start(sw->fh, start_ts, dts, follower, num_crop, sw->im_w, sw->im_h, sw->fps_id, sw->comp_num);
 	for (i = 0; i < num_crop; i++)
-		write_pcs_start_obj(sw->fh, i, i, crops[i].x, crops[i].y);
+		write_pcs_start_obj(sw->fh, sw->picture_offset + i, in_window[i], crops[i].x, crops[i].y);
 
 	/* Write WDS */
-	ts = start_ts - window_ts;
-	write_wds(sw->fh, ts, dts, num_crop);
-	for (i = 0; i < num_crop; i++)
-		write_wds_obj(sw->fh, i, crops[i].w, crops[i].h, crops[i].x, crops[i].y);
+	ts = start_ts - window_ts; /* Can be very slightly off, possible rounding error (FIXME: fixed?) */
+	write_wds(sw->fh, ts, dts, sw->window_num);
+	for (i = 0; i < sw->window_num; i++)
+		write_wds_obj(sw->fh, i, sw->windows[i].w, sw->windows[i].h, sw->windows[i].x, sw->windows[i].y);
 
 	/* Write palette */
-	write_palette(sw->fh, dts, 0, pal, sw->colorspace);
+	write_palette(sw->fh, dts, sw->palette_offset, pal, sw->colorspace);
 
 	/* Write image data */
-	ts = start_ts - frame_ts + window_ts;
 	for (i = 0; i < num_crop; i++)
 	{
-		rle_len = 0;
-		rle = rl_encode(im, sw->im_w, sw->im_h, crops[i], &rle_len);
 		if (num_crop == 1)
 		{
-			im_ts = ts;
-			im_dts = dts;
+			if (new_composition)
+				im_ts = start_ts - frame_ts + window_ts; /* This one can be off a bit. (FIXME) */
+			else
+				im_ts = start_ts - window_ts;
 		}
 		else if (i == 0)
 		{
-			im_ts = ts;
-			im_dts = im_ts - window_ts;
+			if (new_composition)
+				im_ts = start_ts - frame_ts + window_ts_list[0] - later_window; /* This one can be off a bit, ~5/90000s was observed (FIXME: fixed?) */
+			else
+				im_ts = start_ts - later_window - decode_ts_list[1];
 		}
 		else
 		{
-			im_dts = ts;
-			im_ts = im_dts + window_ts; 
+			if (new_composition)
+			{
+				im_ts = start_ts - frame_ts + window_ts; /* Can be slightly off, possible rounding error (FIXME: fixed?) */
+				dts = start_ts - frame_ts + window_ts_list[0] - later_window;
+			}
+			else
+			{
+				im_ts = start_ts - later_window;
+				dts = start_ts - later_window - decode_ts_list[1];
+			}
 		}
-		write_image(sw->fh, im_ts, im_dts, i, crops[i].w, crops[i].h, rle, rle_len);
-		(sw->picture)++;
-		sw->buffer += crops[i].w * crops[i].h * 4;
-		free(rle);
+		write_image(sw->fh, im_ts, dts, sw->picture_offset + i, crops[i].w, crops[i].h, rle[i], rle_len[i]);
 	}
 
 	/* Write marker */
-	write_marker(sw->fh, ts);
+	write_marker(sw->fh, im_ts);
 
 	/* Remember data for creation of composition end */
-	sw->last.end_ts = end_ts;
-	sw->last.window_ts = window_ts;
-	sw->last.num_crop = num_crop;
-	for (i = 0; i < 2; i++)
+	sw->last_end_ts = end_ts;
+	sw->last_window_ts = window_ts;
+}
+
+void write_composition (sup_writer_t *sw)
+{
+	rect_t *rects;
+	subtitle_info_t *si;
+	int last_num_crop = 0;
+	rect_t last_crops[2];
+	int new_composition = 1;
+	int si_rects = 0;
+	int ts, dts;
+	int i;
+
+	/* Only write anything if there is a non-empty composition */
+	if (!sw->non_new)
+		return;
+
+	/* Count subtitles */
+	si = si_list_first(sw->sil);
+	while (si != NULL)
 	{
-		sw->last.crops[i].x = crops[i].x;
-		sw->last.crops[i].y = crops[i].y;
-		sw->last.crops[i].w = crops[i].w;
-		sw->last.crops[i].h = crops[i].h;
+		si_rects += si->num_crop;
+		si = si_list_next(sw->sil);
 	}
+
+	/* Gather crop rects. */
+	rects = malloc(si_rects * sizeof(rect_t));
+	si = si_list_first(sw->sil);
+	si_rects = 0;
+	while (si != NULL)
+	{
+		for (i = 0; i < si->num_crop; i++)
+			rects[si_rects++] = si->crops[i];
+		si = si_list_next(sw->sil);
+	}
+
+	/* Calculate windows */
+	sw->window_num = find_windows(rects, si_rects, sw->windows);
+
+	if (!sw->window_num)
+	{
+		fprintf(stderr, "Warning: WDS failure, skipping.\n");
+		return;
+	}
+
+	/* Write subtitles */
+	si = si_list_first(sw->sil);
+	while (si != NULL)
+	{
+		if (!new_composition && (last_num_crop != si->num_crop || memcmp(last_crops, si->crops, MIN(last_num_crop, si->num_crop) * sizeof(rect_t))))
+		{
+			(sw->comp_num)++;
+			(sw->palette_offset)++;
+			(sw->picture_offset) += last_num_crop;
+		}
+		last_num_crop = si->num_crop;
+		memcpy(last_crops, si->crops, si->num_crop * sizeof(rect_t));
+		write_subtitle(sw, si->rle, si->rle_len, si->num_crop, si->crops, si->pal, si->start, si->end, new_composition);
+		new_composition = 0;
+		si_list_delete(sw->sil);
+		destroy_si(si);
+		si = si_list_get(sw->sil);
+	}
+
+	/* Write PCSE */
+	dts = sw->last_end_ts - sw->last_window_ts - 1;
+	write_pcs_end(sw->fh, sw->last_end_ts, dts, sw->im_w, sw->im_h, sw->fps_id, ++(sw->comp_num));
+
+	/* Write WDS */
+	ts = sw->last_end_ts - sw->last_window_ts;
+	write_wds(sw->fh, ts, dts, sw->window_num);
+	for (i = 0; i < sw->window_num; i++)
+		write_wds_obj(sw->fh, i, sw->windows[i].w, sw->windows[i].h, sw->windows[i].x, sw->windows[i].y);
+
+	/* Write marker */
+	write_marker(sw->fh, dts);
+
+	/* Cleanup */
+	free(rects);
+
+	/* New composition */
+	(sw->comp_num)++;
+
+	/* Reset picture and palette count, new buffer. */
+	sw->objects = 0;
+	sw->palettes = 0;
+	sw->palette_offset = 0;
+	sw->picture_offset = 0;
+	sw->buffer = 0;
+}
+
+subtitle_info_t *collect_si (sup_writer_t *sw, uint8_t *im, int num_crop, rect_t *crops, uint32_t *pal, int start, int end)
+{
+	subtitle_info_t *si = malloc(sizeof(subtitle_info_t));
+	int i;
+
+	si->start = start;
+	si->end = end;
+	si->num_crop = num_crop;
+	for (i = 0; i < num_crop; i++)
+	{
+		si->crops[i].w = crops[i].w;
+		si->crops[i].h = crops[i].h;
+		si->crops[i].x = crops[i].x;
+		si->crops[i].y = crops[i].y;
+		si->rle[i] = rl_encode(im, sw->im_w, sw->im_h, si->crops[i], &(si->rle_len[i]));
+	}
+	memcpy(si->pal, pal, 256 * sizeof(uint32_t));
+
+	return si;
+}
+
+void close_sup_writer (sup_writer_t *sw)
+{
+	write_composition(sw);
+
+	si_list_first(sw->sil);
+	while(!si_list_empty(sw->sil))
+	{
+		destroy_si(si_list_first(sw->sil));
+		si_list_delete(sw->sil);
+	}
+	si_list_destroy(sw->sil);
+
+	fclose(sw->fh);
+	free(sw);
+}
+
+IMPLEMENT_LIST(si, subtitle_info_t)
+
+void write_sup (sup_writer_t *sw, uint8_t *im, int num_crop, rect_t *crops, uint32_t *pal, int start, int end)
+{
+	rect_t tmp;
+	int buffer_increase;
+	int i;
+
+	buffer_increase = 0;
+	for (i = 0; i < num_crop; i++)
+		buffer_increase += crops[i].w * crops[i].h + 16;
+	/* Disabled some conditions for now. */
+	if (sw->non_new && ((start > sw->end + 1) || /*(sw->buffer + buffer_increase >= 4 * 1024 * 1024) ||*/ (sw->objects + num_crop > 64) /*-|| (sw->palettes + 1 > 8)*/))
+	{
+#		if DEBUG != 0
+#		warning "DEBUG enabled."
+			printf("Starting new composition ");
+			if (start > sw->end)
+				printf("due to time difference. %u > %u + 1\n", start, sw->end);
+			else if (sw->buffer + buffer_increase >= 4 * 1024 * 1024)
+				printf("due to buffer overflow. %u + %u = %u > %u\n", sw->buffer, buffer_increase, sw->buffer + buffer_increase, 4*1024*1024);
+			else if (sw->objects + num_crop > 64)
+				printf("due to number of composition objects. %u + %u = %u > %u\n", sw->objects, num_crop, sw->picture + num_crop, 64);
+			else if (sw->palette + 1 > 8)
+				printf("due to number of palettes. %u + %u = %u > %u\n", sw->palette, 1, sw->palette + 1, 8);
+			else
+				printf("for unknown reasons.\n");
+#		endif
+		write_composition(sw);
+	}
+	sw->non_new = 1;
+	sw->end = end;
+	sw->buffer += buffer_increase;
+	sw->objects += num_crop; /* FIXME? */
+	(sw->palettes)++; /* FIXME: This is dumb. */
+
+	if (num_crop > 1)
+	{
+		/* Let's order them, so the one closer to 0/0 is the second. */
+		if (crops[0].y < crops[1].y || (crops[0].y == crops[1].y && crops[0].x < crops[1].x))
+		{
+			tmp = crops[0];
+			crops[0] = crops[1];
+			crops[1] = tmp;
+		}
+	}
+
+	si_list_insert_after(sw->sil, collect_si(sw, im, num_crop, crops, pal, start, end));
 }
 
